#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>

#include "pcplib.h"

#define dprintf(A...) fprintf(stderr , ##A)

static const char STX = '\002';
static const char ETX = '\003';
static const char SEP = ':';

#define STX_STR_ETX(A) "\002" #A "\003"

struct proj_status_t {
    int power;
    int freeze;
    int shutter;
    int lampPower;
    int autoKeystone;
    int autoRGB;
    int autoYPbPr;
    int volume;
    int brightness;
    int color;
    int contrast;
    int sharpness;
    int tint;
    int backgroundColor;
    int pictureMode;
    int lampOnTime;
    int lampOnCondition;
    char language[4];    
};

static struct proj_status_t proj;

const char * serialSpeedVect[] = {
    "4800", "9600", "19200", "38400", NULL
};

static const int serialSpeedCode[] = {
    B4800, B9600, B19200, B38400
};

extern char * defaultSerialDevice;
static char * serialDevice;
static int serialPort;
static int serialSpeed;
static int defaultTimeout = 3;

#define DEFAULT_SERIAL_DEVICE "/dev/ttyS0"
#define ALTERNATE_SERIAL_DEVICE "/dev/ttyS1"
#define DEFAULT_SERIAL_SPEED  9600

#define GETF(A) int pcp_get_ ##A (void) { return proj. ##A ; }

GETF(power)
GETF(freeze)
GETF(shutter)
GETF(lampPower)
GETF(autoKeystone)
GETF(autoRGB)
GETF(autoYPbPr)
GETF(volume)
GETF(brightness)
GETF(color)
GETF(contrast)
GETF(sharpness)
GETF(tint)
GETF(backgroundColor)
GETF(pictureMode)
GETF(lampOnTime)
GETF(lampOnCondition)

#undef GETF

static int is_error(char res[])
{
    dprintf("Response error: \"%s\"\n", res);
    dprintf("(%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d)\n",
            res[0], res[1], res[2], res[3], res[4], res[5], res[6],
            res[7], res[8], res[9], res[10], res[11], res[12], res[13],
            res[14], res[15]); 
    if (!strcmp(res,STX_STR_ETX(ER401)))
        return PCP_ER401;
    if (!strcmp(res,STX_STR_ETX(ER402)))
        return PCP_ER402;
    return PCP_ERESP;
}

static void build_command(char *code, char *arg, char *res)
{
    *res++ = STX;
    while (*code)
        *res++ = *code++;
    if (arg) {
        *res++ = SEP;
        while (*arg)
            *res++ = *arg++;
    }
    *res++ = ETX;
    *res = '\0';
}

static void send_command(char * code)
{
    int i = 0, l = strlen(code);
    
    if (serialPort == -1)
        return;
    
    tcflush(serialPort, TCIOFLUSH); /* flush I/O data still pending */
    
    do {
        i += write(serialPort, code, l-i);
        if (i == -1) {
            perror(serialDevice);
            exit(1);
        }
    } while (l < i);
    
    tcdrain(serialPort); /* wait until data has been written */
}

static int get_response(char *resp, int secs)
{
    int l;

    if (serialPort == -1)
        return PCP_NOPORT;
        
    *resp = '\0';
    alarm(secs);
    do {
        l = read(serialPort, resp, 1);
    } while (l>=0 && *resp++ != ETX);
    alarm(0);
    *resp = '\0';
    return (l >= 0 ? PCP_OK : (errno == EINTR ? PCP_TIMEOUT : PCP_ERESP));
}

#define _Finv(A,B) \
int pcp_ ##B## _command(void) \
{ \
    int r; \
    char *cmd, res[16]; \
    cmd = STX_STR_ETX(A); \
    send_command(cmd); \
    if ((r = get_response(res, defaultTimeout)) != PCP_OK) \
        return r; \
    if (!strcmp(cmd,res)) { \
        proj. ##B = !proj. ##B; \
        return PCP_OK; \
    } \
    return is_error(res); \
}

#define _F01(A,B) \
int pcp_ ##B## _command(int onoff) \
{ \
    int r; \
    char cmd[16], res[16], arg[2]; \
    arg[0] = '0'+onoff; \
    arg[1] = '\0'; \
    build_command(#A, arg, cmd); \
    send_command(cmd); \
    if ((r = get_response(res, defaultTimeout)) != PCP_OK) \
        return r; \
    if (!strcmp(cmd,res) ) { \
        proj. ##B = onoff; \
        return PCP_OK; \
    } \
    return is_error(res); \
}


#define _Fset64(A,B) \
int pcp_ ##B## _command(int value) \
{ \
    int r; \
    char cmd[16], res[16], arg[4]; \
    arg[0] = '0'; \
    arg[1] = '0' + value / 10; \
    arg[2] = '0' + value % 10; \
    arg[3] = '\0'; \
    build_command(#A, arg, cmd); \
    send_command(cmd); \
    if ((r = get_response(res, defaultTimeout)) != PCP_OK) \
        return r; \
    if (!strcmp(cmd,res) ) { \
        proj. ##B = value; \
        return PCP_OK; \
    } \
    return is_error(res); \
}


int pcp_pictureMode_command(int onoff)
{
    int r;
    char cmd[16], res[16], *arg;
    arg = (onoff ? "DYN" : "NAT"); 
    build_command("VPM", arg, cmd);
    send_command(cmd);
    if ((r = get_response(res, defaultTimeout)) != PCP_OK)
        return r;
    if (!strcmp(cmd,res) ) {
        proj.pictureMode = onoff;
        return PCP_OK;
    }
    return is_error(res);
}


int pcp_powerOff_command(void)
{
    int r;
    char *cmd, res[16];
    
    cmd = STX_STR_ETX(POF);
    send_command(cmd);
    if ((r = get_response(res, defaultTimeout)) != PCP_OK)
        return r;
    if (!strcmp(cmd,res)) {
        proj.power = 0;
        return r;
    }
    return is_error(res);
}

int pcp_powerOn_command(void)
{
    int r;
    char *cmd, res[16];
    
    cmd = STX_STR_ETX(PON);
    send_command(cmd);
    
    if ((r = get_response(res, defaultTimeout)) != PCP_OK)
        return r;
    
    if (!strcmp(cmd,res)) {
        proj.power = 1;
        return r;
    }
    return is_error(res);
}

_F01(OFZ,freeze)
_F01(OLP,lampPower)
_F01(OBC,backgroundColor)
_F01(OAK,autoKeystone)
_F01(OAR,autoRGB)
_F01(OAY,autoYPbPr)
_Finv(OSH,shutter)
_Fset64(VCN,contrast)
_Fset64(VBR,brightness)
_Fset64(VCO,color)
_Fset64(VSR,sharpness)
_Fset64(VTN,tint)
_Fset64(AVL,volume)


#undef _Finv
#undef _Fs0
#undef _Fset64



#define QUERY01(A,B) \
static int query_##A(void) \
{ \
    int r; \
    char res[16]; \
    send_command(STX_STR_ETX(B)); \
    if ((r = get_response(res, defaultTimeout)) != PCP_OK) \
        return r; \
    if (res[0] == STX && res[2] == ETX && \
        (res[1] == '0' || res[1] == '1')) \
        return res[1]-'0'; \
    return is_error(res); \
}

#define QUERY03(A,B) \
static int query_##A(void) \
{ \
    int r; \
    char res[16]; \
    send_command(STX_STR_ETX(B)); \
    if ((r = get_response(res, defaultTimeout)) != PCP_OK) \
        return r; \
    if (res[0] == STX && res[2] == ETX && \
        res[1] >= '0' && res[1] <= '3') \
        return res[1]-'0'; \
    return is_error(res); \
}

#define QUERY063(A,B) \
static int query_##A(void) \
{ \
    int r; \
    char res[16]; \
    send_command(STX_STR_ETX(B)); \
    if ((r = get_response(res, defaultTimeout)) != PCP_OK) \
        return r; \
    if (res[0] == STX && res[1] == '0' && res[4] == ETX && \
       res[2] >= '0' && res[2] <= '6' && res[3] >= '0' && res[3] <= '9') \
        return (res[2]-'0')*10+(res[3]-'0'); \
    return is_error(res); \
}

QUERY01(autoKeystone,    QAK)
QUERY01(autoRGB,         QAR)
QUERY01(autoYPbPr,       QAY)
QUERY01(backgroundColor, QBC)
QUERY01(lampPower,       QLP)
QUERY01(shutter,         QSH)
QUERY01(freeze,          QFZ)

QUERY03(lampOnCondition, Q$S)

QUERY063(volume,         QAV)
QUERY063(brightness,     QVB)
QUERY063(color,          QVC)
QUERY063(contrast,       QVR)
QUERY063(sharpness,      QVS)
QUERY063(tint,           QVT)

#undef QUERY01
#undef QUERY03
#undef QUERY063

static void query_language(char b[])
{
    int r;
    char res[16];
    
    send_command(STX_STR_ETX(QLG));
    r = get_response(res, defaultTimeout);
    if (r != PCP_OK || res[0] != STX || res[4] != ETX) 
        b[0] = '\0';
    else { 
      *b++ = res[1];
      *b++ = res[2];
      *b++ = res[3];
      *b = '\0';
    }
}

static int query_pictureMode(void)
{
    int r;
    char res[16];
    
    send_command(STX_STR_ETX(QPM));
    if ((r = get_response(res, defaultTimeout)) != PCP_OK)
        return r;
    if (res[0] == STX && res[4] == ETX) {
        if (res[1] == 'N' && res[2] == 'A' && res[3] == 'T')
            return 0;
        else if (res[1] == 'D' && res[2] == 'Y' && res[3] == 'N')
            return 1;
    }
    return is_error(res);
}


static int query_lampOnTime(void)
{
    int r;
    char res[16];
    send_command(STX_STR_ETX(Q$L));
    if ((r = get_response(res, defaultTimeout)) != PCP_OK)
        return r;
    if (res[0] == STX && res[5] == ETX &&
        res[1] >= '0' && res[1] <= '9' && res[2] >= '0' && res[2] <= '9' &&
        res[3] >= '0' && res[3] <= '9' && res[4] >= '0' && res[4] <= '9')
        return (res[1]-'0')*1000+(res[2]-'0')*100+(res[3]-'0')*10+(res[4]-'0');
    return is_error(res);    
}

int query_power(void)
{
    int r;
    char res[16];
    
    send_command(STX_STR_ETX(QPW));
    if ((r = get_response(res, defaultTimeout)) != PCP_OK)
        return r;
    if (res[0]==STX && res[1]=='0' && res[2]=='0' && 
        res[4]==ETX && (res[3] == '0' || res[3] == '1'))
        return res[3]-'0';
    return is_error(res);
}


#define PROBE(A,B) \
    if ((r = query_ ##A## ()) >= 0 && \
            pcp_ ##A## _command( B ) == PCP_OK && \
            (p = query_ ##A## ()) >= 0 && \
            p != r && \
            pcp_ ##A## _command( r ) == PCP_OK) \
        proj. ##A = r; \
    else { \
        proj. ##A = -1; \
        /* dprintf("[W] %s not supported\n", #A); */ \
    }
    
void query_projector_status(void)
{
    int r, p;
    
    
restart:
        r = query_power();
        if (r < 0) /* it was r<=0, but it doesn't look correct to me */
            goto restart;
        if (r == 0) {
            pcp_powerOn_command();
            sleep(defaultTimeout);
            goto restart;
        }
        
        r = query_lampOnCondition();
        
        switch (r) {
            case 0: /* in stand-by */
                fprintf(stderr, "Unexpected state: QPW=1, Q$S=0\n");
                exit(1);
            case 1: /* warming up */
                sleep(defaultTimeout);
                goto restart;
            case 2: /* lamp on */
                break;
            case 3: /* cooling down */
                /* the projector was cooling the lamp when the powerOn
                   command was issued: we must wait till the cooling phase
                   terminates */
                sleep(defaultTimeout);
                goto restart;
        }
    
    r = query_lampOnCondition();
    proj.lampOnCondition = (r >= 0 ? r : -1);

    if (proj.lampOnCondition != 2) {
        fprintf(stderr, "Cannot initialize the projector\n");
        return;
    }
    
    r = query_shutter();
    proj.shutter = (r >= 0 ? r : -1);
    if (r == 1) {
        fprintf(stderr, "Projector found in shutter mode\n");
        pcp_shutter_command();
    }
    

    PROBE(autoKeystone, !r)
    PROBE(autoRGB, !r)
    PROBE(autoYPbPr, !r) 
    PROBE(backgroundColor, !r)
    PROBE(freeze, !r)
    PROBE(volume, (r!=0 ? 0 : 10))
    PROBE(brightness, (r!=0 ? 0 : 10))
    PROBE(color, (r!=0 ? 0 : 10))
    PROBE(contrast, (r!=0 ? 0 : 10))
    PROBE(sharpness, (r!=0 ? 0 : 10))
    PROBE(tint, (r!=0 ? 0 : 10))
    PROBE(pictureMode, !r)
    PROBE(lampPower, !r)

        
    r = query_lampOnTime();
    proj.lampOnTime = (r >= 0 ? r : -1);
    
    query_language(proj.language);           
 }



int initSerialDevice(char *log)
{
    struct termios tio;

    serialPort = open(serialDevice, O_RDWR);
    if (serialPort == -1)
        goto exit_l;
            
    if (tcgetattr(serialPort, &tio))
        goto exit_l;
        
    cfmakeraw(&tio);
    
    if (  cfsetospeed(&tio, serialSpeed)        ||
          cfsetispeed(&tio, serialSpeed)        ||
          tcsetattr(serialPort, TCSAFLUSH, &tio)      ) {
exit_l:
        if (log)
            strncpy(log, strerror(errno), 128);
        return 1;
    }
    return 0;
}



int pcp_set_serialDevice(const char * dev, char *log)
{
    if (serialPort != -1 && close(serialPort)) {
        if (log) strncpy(log, strerror(errno), 128);
        return 1;
    }
    
    free(serialDevice);

    if (!strcmp(dev,COMBO_SERIAL_DEVICE_0))
        serialDevice = strdup(DEFAULT_SERIAL_DEVICE);
    else if (!strcmp(dev, COMBO_SERIAL_DEVICE_1))
        serialDevice = strdup(ALTERNATE_SERIAL_DEVICE);
    else serialDevice = strdup(dev);
    
    return initSerialDevice(log);    
}


int pcp_set_serialSpeed(int idx, char * log)
{
    
    if (idx < 0 || idx > 3) {
        strncpy(log, "Invalid speed rate", 128);
        return 1;
    }
    
    serialSpeed = serialSpeedCode[idx];
    return initSerialDevice(log);
}

void pcp_set_timeoutDefault(int v)
{   
    defaultTimeout = v;
}

static void exithook(void)
{
    if (proj.power == 1) 
        dprintf("Warning: projector left in non-standby mode\n");
}

static void sigalrm(int unused __attribute__((unused))) { /* do nothing */ };

int pcplib_init(void)
{
    struct sigaction sa;
    
    sa.sa_handler = sigalrm;
    sa.sa_flags = 0;
    
    if (sigaction(SIGALRM, &sa, NULL)) {
        fprintf(stderr, "Cannot install signal handler\n");
        exit(1);
    }


    if (!defaultSerialDevice)
        serialDevice = strdup(DEFAULT_SERIAL_DEVICE);
    else
        serialDevice = strdup(defaultSerialDevice);

    serialSpeed = serialSpeedCode[serialSpeedDefault];

    initSerialDevice(NULL);
           
    atexit(exithook);
    
    proj.power = -1;

    /* the very first command awakes the projector if
       it is muted (shutter=1) --- any command beside OSH 
       returns ER401 --- but we don't want to issue an OSH,
       or the active projector will mute on startup */
    
    query_lampOnCondition();    
    
    return 0;
}
